@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2019 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 /*
@@ -30,32 +30,39 @@
 #include "jbig2_priv.h"
 #include "jbig2_image.h"
 
+#if !defined (INT32_MAX)
+#define INT32_MAX  0x7fffffff
+#endif
+
 /* allocate a Jbig2Image structure and its associated bitmap */
 Jbig2Image *
-jbig2_image_new(Jbig2Ctx *ctx, int width, int height)
+jbig2_image_new(Jbig2Ctx *ctx, uint32_t width, uint32_t height)
 {
     Jbig2Image *image;
-    int stride;
-    int64_t check;
+    uint32_t stride;
+
+    if (width == 0 || height == 0) {
+        jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "failed to create zero sized image");
+        return NULL;
+    }
 
     image = jbig2_new(ctx, Jbig2Image, 1);
     if (image == NULL) {
-        jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "could not allocate image structure in jbig2_image_new");
+        jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "failed to allocate image");
         return NULL;
     }
 
     stride = ((width - 1) >> 3) + 1;    /* generate a byte-aligned stride */
+
     /* check for integer multiplication overflow */
-    check = ((int64_t) stride) * ((int64_t) height);
-    if (check != (int)check) {
-        jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "integer multiplication overflow from stride(%d)*height(%d)", stride, height);
+    if (height > (INT32_MAX / stride)) {
+        jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "integer multiplication overflow (stride=%u, height=%u)", stride, height);
         jbig2_free(ctx->allocator, image);
         return NULL;
     }
-    /* Add 1 to accept runs that exceed image width and clamped to width+1 */
-    image->data = jbig2_new(ctx, uint8_t, (int)check + 1);
+    image->data = jbig2_new(ctx, uint8_t, (size_t) height * stride);
     if (image->data == NULL) {
-        jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "could not allocate image data buffer! [stride(%d)*height(%d) bytes]", stride, height);
+        jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "failed to allocate image data buffer (stride=%u, height=%u)", stride, height);
         jbig2_free(ctx->allocator, image);
         return NULL;
     }
@@ -68,9 +75,9 @@ jbig2_image_new(Jbig2Ctx *ctx, int width, int height)
     return image;
 }
 
-/* clone an image pointer by bumping its reference count */
+/* bump the reference count for an image pointer */
 Jbig2Image *
-jbig2_image_clone(Jbig2Ctx *ctx, Jbig2Image *image)
+jbig2_image_reference(Jbig2Ctx *ctx, Jbig2Image *image)
 {
     if (image)
         image->refcount++;
@@ -84,7 +91,7 @@ jbig2_image_release(Jbig2Ctx *ctx, Jbig2Image *image)
     if (image == NULL)
         return;
     image->refcount--;
-    if (!image->refcount)
+    if (image->refcount == 0)
         jbig2_image_free(ctx, image);
 }
 
@@ -92,40 +99,66 @@ jbig2_image_release(Jbig2Ctx *ctx, Jbig2Image *image)
 void
 jbig2_image_free(Jbig2Ctx *ctx, Jbig2Image *image)
 {
-    if (image)
+    if (image != NULL) {
         jbig2_free(ctx->allocator, image->data);
-    jbig2_free(ctx->allocator, image);
+        jbig2_free(ctx->allocator, image);
+    }
 }
 
 /* resize a Jbig2Image */
 Jbig2Image *
-jbig2_image_resize(Jbig2Ctx *ctx, Jbig2Image *image, int width, int height)
+jbig2_image_resize(Jbig2Ctx *ctx, Jbig2Image *image, uint32_t width, uint32_t height, int value)
 {
     if (width == image->width) {
         /* check for integer multiplication overflow */
-        int64_t check = ((int64_t) image->stride) * ((int64_t) height);
-
-        if (check != (int)check) {
-            jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "integer multiplication overflow during resize stride(%d)*height(%d)", image->stride, height);
+        if (image->height > (INT32_MAX / image->stride)) {
+            jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "integer multiplication overflow during resize (stride=%u, height=%u)", image->stride, height);
             return NULL;
         }
         /* use the same stride, just change the length */
-        image->data = jbig2_renew(ctx, image->data, uint8_t, (int)check);
+        image->data = jbig2_renew(ctx, image->data, uint8_t, (size_t) height * image->stride);
         if (image->data == NULL) {
-            jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "could not resize image buffer!");
+            jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "failed to reallocate image");
             return NULL;
         }
         if (height > image->height) {
-            memset(image->data + image->height * image->stride, 0, (height - image->height) * image->stride);
+            const uint8_t fill = value ? 0xFF : 0x00;
+            memset(image->data + (size_t) image->height * image->stride, fill, ((size_t) height - image->height) * image->stride);
         }
         image->height = height;
 
     } else {
-        /* we must allocate a new image buffer and copy */
-        jbig2_error(ctx, JBIG2_SEVERITY_WARNING, -1, "jbig2_image_resize called with a different width (NYI)");
+        Jbig2Image *newimage;
+        int code;
+
+        /* Unoptimized implementation, but it works. */
+
+        newimage = jbig2_image_new(ctx, width, height);
+        if (newimage == NULL) {
+            jbig2_error(ctx, JBIG2_SEVERITY_WARNING, -1, "failed to allocate resized image");
+            return NULL;
+        }
+        jbig2_image_clear(ctx, newimage, value);
+
+        code = jbig2_image_compose(ctx, newimage, image, 0, 0, JBIG2_COMPOSE_REPLACE);
+        if (code < 0) {
+            jbig2_error(ctx, JBIG2_SEVERITY_WARNING, -1, "failed to compose image buffers when resizing");
+            jbig2_image_release(ctx, newimage);
+            return NULL;
+        }
+
+        /* if refcount > 1 the original image, its pointer must
+        be kept, so simply replaces its innards, and throw away
+        the empty new image shell. */
+        jbig2_free(ctx->allocator, image->data);
+        image->width = newimage->width;
+        image->height = newimage->height;
+        image->stride = newimage->stride;
+        image->data = newimage->data;
+        jbig2_free(ctx->allocator, newimage);
     }
 
-    return NULL;
+    return image;
 }
 
 /* composite one jbig2_image onto another
@@ -133,27 +166,41 @@ jbig2_image_resize(Jbig2Ctx *ctx, Jbig2Image *image, int width, int height)
 static int
 jbig2_image_compose_unopt(Jbig2Ctx *ctx, Jbig2Image *dst, Jbig2Image *src, int x, int y, Jbig2ComposeOp op)
 {
-    int i, j;
-    int sw = src->width;
-    int sh = src->height;
-    int sx = 0;
-    int sy = 0;
+    uint32_t i, j;
+    uint32_t sw = src->width;
+    uint32_t sh = src->height;
+    uint32_t sx = 0;
+    uint32_t sy = 0;
 
     /* clip to the dst image boundaries */
     if (x < 0) {
         sx += -x;
-        sw -= -x;
+        if (sw < (uint32_t) -x)
+            sw = 0;
+        else
+            sw -= -x;
         x = 0;
     }
     if (y < 0) {
         sy += -y;
-        sh -= -y;
+        if (sh < (uint32_t) -y)
+            sh = 0;
+        else
+            sh -= -y;
         y = 0;
     }
-    if (x + sw >= dst->width)
-        sw = dst->width - x;
-    if (y + sh >= dst->height)
-        sh = dst->height - y;
+    if ((uint32_t) x + sw >= dst->width) {
+        if (dst->width >= (uint32_t) x)
+            sw = dst->width - x;
+        else
+            sw = 0;
+    }
+    if ((uint32_t) y + sh >= dst->height) {
+        if (dst->height >= (uint32_t) y)
+            sh = dst->height - y;
+        else
+            sh = 0;
+    }
 
     switch (op) {
     case JBIG2_COMPOSE_OR:
@@ -200,36 +247,64 @@ jbig2_image_compose_unopt(Jbig2Ctx *ctx, Jbig2Image *dst, Jbig2Image *src, int x
 int
 jbig2_image_compose(Jbig2Ctx *ctx, Jbig2Image *dst, Jbig2Image *src, int x, int y, Jbig2ComposeOp op)
 {
-    int i, j;
-    int w, h;
-    int leftbyte, rightbyte;
-    int shift;
+    uint32_t i, j;
+    uint32_t w, h;
+    uint32_t leftbyte, rightbyte;
+    uint32_t shift;
     uint8_t *s, *ss;
     uint8_t *d, *dd;
     uint8_t mask, rightmask;
 
-    if (op != JBIG2_COMPOSE_OR) {
+    if (src == NULL)
+        return 0;
+
+    /* The optimized code for the OR operator below doesn't
+       handle the source image partially placed outside the
+       destination (above and/or to the left). The affected
+       intersection of the destination is computed correctly,
+       however the correct subset of the source image is not
+       chosen. Instead the upper left corner of the source image
+       is always used.
+
+       In the unoptimized version that handles all operators
+       (including OR) the correct subset of the source image is
+       chosen.
+
+       The workaround is to check whether the x/y coordinates to
+       the composition operator are negative and in this case use
+       the unoptimized implementation.
+
+       TODO: Fix the optimized OR implementation if possible. */
+    if (op != JBIG2_COMPOSE_OR || x < 0 || y < 0) {
         /* hand off the the general routine */
         return jbig2_image_compose_unopt(ctx, dst, src, x, y, op);
     }
 
+    /* optimized code for the prevalent OR operator */
+
     /* clip */
     w = src->width;
     h = src->height;
     ss = src->data;
 
     if (x < 0) {
-        w += x;
+        if (w < (uint32_t) -x)
+            w = 0;
+        else
+            w += x;
         x = 0;
     }
     if (y < 0) {
-        h += y;
+        if (h < (uint32_t) -y)
+            h = 0;
+        else
+            h += y;
         y = 0;
     }
-    w = (x + w < dst->width) ? w : dst->width - x;
-    h = (y + h < dst->height) ? h : dst->height - y;
+    w = ((uint32_t) x + w < dst->width) ? w : ((dst->width >= (uint32_t) x) ? dst->width - (uint32_t) x : 0);
+    h = ((uint32_t) y + h < dst->height) ? h : ((dst->height >= (uint32_t) y) ? dst->height - (uint32_t) y : 0);
 #ifdef JBIG2_DEBUG
-    jbig2_error(ctx, JBIG2_SEVERITY_DEBUG, -1, "compositing %dx%d at (%d, %d) after clipping\n", w, h, x, y);
+    jbig2_error(ctx, JBIG2_SEVERITY_DEBUG, -1, "compositing %dx%d at (%d, %d) after clipping", w, h, x, y);
 #endif
 
     /* check for zero clipping region */
@@ -239,24 +314,18 @@ jbig2_image_compose(Jbig2Ctx *ctx, Jbig2Image *dst, Jbig2Image *src, int x, int
 #endif
         return 0;
     }
-#if 0
-    /* special case complete/strip replacement */
-    /* disabled because it's only safe to do when the destination
-       buffer is all-blank. */
-    if ((x == 0) && (w == src->width)) {
-        memcpy(dst->data + y * dst->stride, src->data, h * src->stride);
-        return 0;
-    }
-#endif
 
-    leftbyte = x >> 3;
-    rightbyte = (x + w - 1) >> 3;
+    leftbyte = (uint32_t) x >> 3;
+    rightbyte = ((uint32_t) x + w - 1) >> 3;
     shift = x & 7;
 
     /* general OR case */
     s = ss;
     d = dd = dst->data + y * dst->stride + leftbyte;
-    if (d < dst->data || leftbyte > dst->stride || h * dst->stride < 0 || d - leftbyte + h * dst->stride > dst->data + dst->height * dst->stride) {
+    if (d < dst->data ||
+        leftbyte > dst->stride ||
+        d - leftbyte + (size_t) h * dst->stride > dst->data + (size_t) dst->height * dst->stride ||
+        s - leftbyte + (size_t) (h - 1) * src->stride + rightbyte > src->data + (size_t) src->height * src->stride) {
         return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "preventing heap overflow in jbig2_image_compose");
     }
     if (leftbyte == rightbyte) {
@@ -331,7 +400,7 @@ jbig2_image_get_pixel(Jbig2Image *image, int x, int y)
 }
 
 /* set an individual pixel value in an image */
-int
+void
 jbig2_image_set_pixel(Jbig2Image *image, int x, int y, bool value)
 {
     const int w = image->width;
@@ -340,9 +409,9 @@ jbig2_image_set_pixel(Jbig2Image *image, int x, int y, bool value)
     int bit, byte;
 
     if ((x < 0) || (x >= w))
-        return 0;
+        return;
     if ((y < 0) || (y >= h))
-        return 0;
+        return;
 
     byte = (x >> 3) + y * image->stride;
     bit = 7 - (x & 7);
@@ -350,6 +419,4 @@ jbig2_image_set_pixel(Jbig2Image *image, int x, int y, bool value)
 
     scratch = image->data[byte] & mask;
     image->data[byte] = scratch | (value << bit);
-
-    return 1;
 }
